
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "conf.h"


extern  int             errno;

static  char *          cmd_name;

static	char *	    	base_dir = ".";
static	char *	    	exist_pre = "existing";
static	char *	    	inval_pre = "invalid";

static	int 	    	cmd_pid = 0;
static	int 	    	bdir_len = 1;


void	cerror(const char *msg, int retval) 
{
	fprintf(stderr, "%s: [%s,%d,%d] %s\n",
		cmd_name, msg, retval, errno,
		strerror(errno));
	return;
}


void	build_path(char *buf, const char *name)
{
    	sprintf(buf, "%s/%d_%s.T", base_dir, cmd_pid, name);
}

void	build_existing_path(char *buf, const char *name)
{
    	sprintf(buf, "%s/%s_%s", base_dir, exist_pre, name);
}

void	build_invalid_path(char *buf, const char *name)
{
    	sprintf(buf, "%s/%s/%s", base_dir, inval_pre, name);
}


static const char *_err_name[] = {
	"SUCCESS",      "EPERM",        "ENOENT",       "ESRCH",        
 	"EINTR",	"EIO",  	"ENXIO",	"E2BIG",
 	"ENOEXEC",	"EBADF",	"ECHILD",	"EAGAIN",
 	"ENOMEM",	"EACCES",	"EFAULT",	"ENOTBLK",
 	"EBUSY",	"EEXIST",	"EXDEV",	"ENODEV",
 	"ENOTDIR",	"EISDIR",	"EINVAL",	"ENFILE",
 	"EMFILE",	"ENOTTY",	"ETXTBSY",	"EFBIG",
 	"ENOSPC",	"ESPIPE",	"EROFS",	"EMLINK",
 	"EPIPE",	"EDOM", 	"ERANGE",	"EDEADLK",
 	"ENAMETOOLONG", "ENOLCK",	"ENOSYS",	"ENOTEMPTY",
 	"ELOOP",	"[41]", 	"ENOMSG",	"EIDRM",
 	"ECHRNG",	"EL2NSYNC",	"EL3HLT",	"EL3RST",
 	"ELNRNG",	"EUNATCH",	"ENOCSI",	"EL2HLT",
 	"EBADE",	"EBADR",	"EXFULL",	"ENOANO",
 	"EBADRQC",	"EBADSLT",	"[58]", 	"EBFONT",
 	"ENOSTR",	"ENODATA",	"ETIME",	"ENOSR",
 	"ENONET",	"ENOPKG",	"EREMOTE",	"ENOLINK",
 	"EADV", 	"ESRMNT",	"ECOMM",	"EPROTO",
 	"EMULTIHOP",	"EDOTDOT",	"EBADMSG",	"EOVERFLOW",
 	"ENOTUNIQ",	"EBADFD",	"EREMCHG",	"ELIBACC",
 	"ELIBBAD",	"ELIBSCN",	"ELIBMAX",	"ELIBEXEC",
 	"EILSEQ",	"ERESTART",	"ESTRPIPE",	"EUSERS",
 	"ENOTSOCK",	"EDESTADDRREQ", "EMSGSIZE",	"EPROTOTYPE",
 	"ENOPROTOOPT",  "EPROTONOSUPPORT",	"ESOCKTNOSUPPORT",	"EOPNOTSUPP",
 	"EPFNOSUPPORT", "EAFNOSUPPORT", "EADDRINUSE",	"EADDRNOTAVAIL",
 	"ENETDOWN",	"ENETUNREACH",  "ENETRESET",	"ECONNABORTED",
 	"ECONNRESET",	"ENOBUFS",	"EISCONN",	"ENOTCONN",
 	"ESHUTDOWN",	"ETOOMANYREFS", "ETIMEDOUT",	"ECONNREFUSED",
 	"EHOSTDOWN",	"EHOSTUNREACH", "EALREADY",	"EINPROGRESS",
 	"ESTALE",	"EUCLEAN",	"ENOTNAM",	"ENAVAIL",
 	"EISNAM",	"EREMOTEIO",	"EDQUOT",	"ENOMEDIUM",
 	"EMEDIUMTYPE",  "[?]"
       };


static inline const char *err_name(int errno)
{
	int index = ((errno <= 0) && (errno > -125)) ? -errno : 125;
	return _err_name[index];
}


void	report(const char *title, const char *file, int retval)
{
    	if (retval > 0) retval=0;
    	fprintf(stdout, "%-14.14s\t%-32.32s %-20.20s [%s]\n", 
	    err_name(-errno), title, 
	    file + bdir_len + 1, strerror(errno));
}



int 	main(int argc,	char *argv[])
{
//  	struct sigaction action;
        extern int optind;
        extern char *optarg;
	char buffer[MAXPATHLEN]; 
	char second[MAXPATHLEN]; 
        char c, errflg = 0;
	int r,f;

        cmd_name = argv[0];
        while ((c = getopt(argc, argv, OPTIONS)) != EOF) {
            switch (c) {
            case 'h':
                fprintf(stderr,
                    "This is %s " VERSION "\n"
                    "options are:\n"
                    "-h        print this help message\n"
                    "-d <dir>  test base directory\n"
                    "-e <pre>  existing wossname prefix\n"
                    ,cmd_name);
                exit(0);
                break;
            case 'd':
                base_dir = optarg;
                break;
            case '?':
            default:
                errflg++;
                break;
            }
        }

        if (errflg) {
            fprintf(stderr, 
                "Usage: %s -[" OPTIONS "]\n"
                "%s -h for help.\n",
                cmd_name, cmd_name);
            exit(2);
        }


    	cmd_pid = getpid();
	bdir_len = strlen(base_dir);


#define XX_SEP(n)   	    	    	    	    	\
    	fprintf(stdout, "--------------\t%s\n", n);

    	/* open calls ... */

#define XE_OPEN(n,f) 	    	    	    	    	\
	errno = 0;  	    	    	    	    	\
    	build_existing_path(buffer, n);    	    	\
	r = open(buffer, f);	    	    	    	\
	report("open " n " " #f, buffer, r);	    	\
	if (r>0) close(r);

#define XN_OPEN(n,f) 	    	    	    	    	\
	errno = 0;  	    	    	    	    	\
    	build_path(buffer, n);    	    	    	\
	r = open(buffer, f, 0);	    	    	    	\
	report("open " n " " #f, buffer, r);	    	\
	if (r>0) close(r);
	
    	{
	    XX_SEP  ("OPEN RDONLY");
	    XE_OPEN ("file",	O_RDONLY);
	    XE_OPEN ("dir",	O_RDONLY);
	    XE_OPEN ("symlink",	O_RDONLY);
	    XE_OPEN ("broken",	O_RDONLY);
	    XE_OPEN ("cdev",	O_RDONLY);
	    XE_OPEN ("bdev",	O_RDONLY);
	    XE_OPEN ("fifo",	O_RDONLY|O_NONBLOCK);
	    
	    XX_SEP  ("OPEN WRONLY");
	    XE_OPEN ("file",	O_WRONLY);
	    XE_OPEN ("dir",	O_WRONLY);
	    XE_OPEN ("symlink",	O_WRONLY);
	    XE_OPEN ("broken",	O_WRONLY);
	    XE_OPEN ("cdev",	O_WRONLY);
	    XE_OPEN ("bdev",	O_WRONLY);
	    XE_OPEN ("fifo",	O_WRONLY|O_NONBLOCK);
	    
	    XX_SEP  ("OPEN RDWR");
	    XE_OPEN ("file",	O_RDWR);
	    XE_OPEN ("dir",	O_RDWR);
	    XE_OPEN ("symlink",	O_RDWR);
	    XE_OPEN ("broken",	O_RDWR);
	    XE_OPEN ("cdev",	O_RDWR);
	    XE_OPEN ("bdev",	O_RDWR);
	    XE_OPEN ("fifo",	O_RDWR|O_NONBLOCK);

	    XX_SEP  ("OPEN CREATE");
	    XN_OPEN ("file",	O_RDONLY);
	    XN_OPEN ("file",	O_WRONLY);
	    XN_OPEN ("file",	O_RDWR);
	    XN_OPEN ("file",	O_CREAT);
	
	}

    	/* mk* calls ... */

#define XE_TREP(t,n,x) 	    	    	    		\
	errno = 0;  	    	    	    	    	\
    	build_existing_path(buffer, n);			\
    	r = x;	    	    	    	    	    	\
	report(t, buffer, r);

#define XN_TREP(t,n,x) 	    	    	    		\
	errno = 0;  	    	    	    	    	\
    	build_path(buffer, n);	    	    	    	\
    	r = x;	    	    	    	    	    	\
	report(t, buffer, r);

#define XI_TREP(t,n,x) 	    	    	    		\
	errno = 0;  	    	    	    	    	\
    	build_invalid_path(buffer, n);	    	       	\
    	r = x;	    	    	    	    	    	\
	report(t, buffer, r);

    	{
	    XX_SEP  ("MKxxx EXIST");
	    XE_TREP ("mkdir", "dir", mkdir(buffer, 0));
	    XE_TREP ("mkfifo",      	"fifo",     mkfifo(buffer, 0));
	    XE_TREP ("mknod S_IFREG", 	"reg", 	    mknod(buffer, S_IFREG, 0));
	    XE_TREP ("mknod S_IFCHR", 	"cdev",     mknod(buffer, S_IFCHR, 0));
	    XE_TREP ("mknod S_IFBLK", 	"bdev",     mknod(buffer, S_IFBLK, 0));
	    XE_TREP ("mknod S_IFIFO", 	"_fifo",    mknod(buffer, S_IFIFO, 0));

	    XX_SEP  ("MKxxx NEW");
	    XN_TREP ("mkdir", 	    	"dir",      mkdir(buffer, 0));
	    XN_TREP ("mkfifo",      	"fifo",     mkfifo(buffer, 0));
	    XN_TREP ("mknod S_IFREG", 	"reg", 	    mknod(buffer, S_IFREG, 0));
	    XN_TREP ("mknod S_IFCHR", 	"cdev",     mknod(buffer, S_IFCHR, 0));
	    XN_TREP ("mknod S_IFBLK", 	"bdev",     mknod(buffer, S_IFBLK, 0));
	    XN_TREP ("mknod S_IFIFO", 	"_fifo",    mknod(buffer, S_IFIFO, 0));

	    XX_SEP  ("MKxxx INVALID");
	    XI_TREP ("mkdir", 	    	"dir",      mkdir(buffer, 0));
	    XI_TREP ("mkfifo",      	"fifo",     mkfifo(buffer, 0));
	    XI_TREP ("mknod S_IFREG", 	"reg", 	    mknod(buffer, S_IFREG, 0));
	    XI_TREP ("mknod S_IFCHR", 	"cdev",     mknod(buffer, S_IFCHR, 0));
	    XI_TREP ("mknod S_IFBLK", 	"bdev",     mknod(buffer, S_IFBLK, 0));
	    XI_TREP ("mknod S_IFIFO", 	"_fifo",    mknod(buffer, S_IFIFO, 0));
	
	}

    	/* ownership/permission stuff ... */

#define XO_TREP(t,n,x) 	    	    	    		\
	errno = 0; r=-1;     	    	    	    	\
    	build_existing_path(buffer, n);			\
    	f = open(buffer, O_RDONLY); 	    	    	\
	if (f>0) r = x;	    	       	    	    	\
	report(t, buffer, r);	    	    	    	\
	if (f>0) close(f);

	{
	    XX_SEP  ("OWNER/PERM");
	    XE_TREP ("chown", 	"file",     	    	chown(buffer, 1,1));
	    XI_TREP ("chown", 	"file",     	    	chown(buffer, 1,1));
	    XO_TREP ("chown", 	"file",     	    	fchown(f, 1,1));
	    XE_TREP ("chmod", 	"file",     	    	chmod(buffer, 0707));
	    XI_TREP ("chmod", 	"file",     	    	chmod(buffer, 0707));
	    XO_TREP ("chmod", 	"file",     	    	fchmod(f, 0707));
    	}


    	/* link & symlink calls ... */

#define XX_TXAB(t,a,b,x) 	    	    	    	\
	errno = 0;  	    	    	    	    	\
    	build_existing_path(buffer, a);	    	    	\
    	build_path(second, b);	    	    	    	\
    	r = x;	    	    	    	    	    	\
	report(t, second, r);

#define XI_TXAB(t,a,b,x) 	    	    	    	\
	errno = 0;  	    	    	    	    	\
    	build_invalid_path(buffer, a);	    	    	\
    	build_path(second, b);	    	    	    	\
    	r = x;	    	    	    	    	    	\
	report(t, second, r);

#define XJ_TXAB(t,a,b,x) 	    	    	    	\
	errno = 0;  	    	    	    	    	\
    	build_existing_path(buffer, a);	    	    	\
    	build_invalid_path(second, b);	     	    	\
    	r = x;	    	    	    	    	    	\
	report(t, second, r);

    	{
	    XX_SEP  ("{,SYM}LINK");
	    XX_TXAB ("link",	"file",     "link",	link(buffer, second));
	    XX_TXAB ("symlink",	"file",     "symlink",	symlink(buffer, second));
	    XI_TXAB ("link",	"file",     "link",	link(buffer, second));
	    XI_TXAB ("symlink",	"file",     "symlink",	symlink(buffer, second));
	    XJ_TXAB ("link",	"file",     "link",	link(buffer, second));
	    XJ_TXAB ("symlink",	"file",     "symlink",	symlink(buffer, second));
	
	}

    	/* other stuff ... */

	{
	    XX_SEP  ("OTHER");
	    XX_TXAB ("rename", 	"reg",	    "_reg",     rename(buffer, second));
	    XI_TXAB ("rename", 	"reg",	    "_reg",     rename(buffer, second));
	    XJ_TXAB ("rename", 	"reg",	    "_reg",     rename(buffer, second));
	    XN_TREP ("rmdir", 	"dir",     	    	rmdir(buffer));
	    XE_TREP ("rmdir", 	"dir",     	    	rmdir(buffer));
	    XI_TREP ("rmdir", 	"dir",     	    	rmdir(buffer));
    	    XN_TREP ("truncate",    	    "file", 	truncate(buffer, 1024));
    	    XE_TREP ("truncate",    	    "file", 	truncate(buffer, 1024));
    	    XI_TREP ("truncate",    	    "file", 	truncate(buffer, 1024));
    	    XO_TREP ("truncate",    	    "file", 	ftruncate(f, 1024));
    	}

    	/* unlink stuff ... */

    	{
	    XX_SEP  ("UNLINK");
	    XN_TREP ("unlink", 	"reg", 	    	    	unlink(buffer));
	    XE_TREP ("unlink", 	"file",     	    	unlink(buffer));
	    XE_TREP ("unlink", 	"bdev",     	    	unlink(buffer));
	    XE_TREP ("unlink", 	"cdev",     	    	unlink(buffer));
	    XE_TREP ("unlink", 	"fifo",     	    	unlink(buffer));
	    XI_TREP ("unlink", 	"file",     	    	unlink(buffer));
	
	}

	return 0;
}
